#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#ifndef _MSC_VER
#include <sys/time.h>
#endif /* _MSC_VER */

#include "windows.h"
#include "gdiplus.h"

#define T_IDENTITY 0x0001
#define T_SCALE    0x0002
#define T_ROTATE   0x0004

#ifdef _MSC_VER
int gettimeofday(struct timeval *tv, void *tz)
{
    ULARGE_INTEGER ull;

    GetSystemTimeAsFileTime((FILETIME *)&ull.u);

    tv->tv_sec = (long)(ull.QuadPart / (ULONGLONG)10000000);
    tv->tv_usec = (long)((ull.QuadPart % (ULONGLONG)10000000) / (ULONGLONG)10);

    return 0;
}
#endif /* _MSC_VER */

static double diff_time(struct timeval *time2, struct timeval *time1)
{
    double t1, t2;

    t1 = (double)time1->tv_sec + (double)time1->tv_usec / 1000000.0;
    t2 = (double)time2->tv_sec + (double)time2->tv_usec / 1000000.0;

    return t2 - t1;
}

static void test_GdipDrawImage(PixelFormat bitmap_fmt, PixelFormat device_fmt,
                               int flags, int n_loop)
{
    struct timeval time1, time2;
    GpStatus status;
    GpBitmap *bitmap, *device_bitmap;
    GpGraphics *graphics;
    int i;

    status = GdipCreateBitmapFromScan0(100, 100, 0, bitmap_fmt, NULL, &bitmap);
    assert(status == Ok);
    status = GdipCreateBitmapFromScan0(100, 100, 0, device_fmt, NULL, &device_bitmap);
    assert(status == Ok);
    status = GdipGetImageGraphicsContext((GpImage *)device_bitmap, &graphics);
    assert(status == Ok);
    status = GdipSetInterpolationMode(graphics, InterpolationModeNearestNeighbor);
    assert(status == Ok);
    status = GdipSetPixelOffsetMode(graphics, PixelOffsetModeHighQuality);
    assert(status == Ok);

    if (flags & T_IDENTITY)
    {
        gettimeofday(&time1, NULL);
        for (i = 0; i < n_loop; i++)
        {
            status = GdipDrawImage(graphics, (GpImage *)bitmap, 0, 0);
            assert(status == Ok);
        }

        gettimeofday(&time2, NULL);
        printf("bitmap format %#x, graphics format %#x (identity): %f seconds\n",
                bitmap_fmt, device_fmt, diff_time(&time2, &time1));
    }

    status = GdipScaleWorldTransform(graphics, 1.5, 3.5, MatrixOrderAppend);
    assert(status == Ok);

    if (flags & T_SCALE)
    {
        gettimeofday(&time1, NULL);
        for (i = 0; i < n_loop; i++)
        {
            status = GdipDrawImage(graphics, (GpImage *)bitmap, 0, 0);
            assert(status == Ok);
        }

        gettimeofday(&time2, NULL);
        printf("bitmap format %#x, graphics format %#x (scale): %f seconds\n",
                bitmap_fmt, device_fmt, diff_time(&time2, &time1));
    }

    status = GdipRotateWorldTransform(graphics, 45.0, MatrixOrderAppend);
    assert(status == Ok);

    if (flags & T_ROTATE)
    {
        gettimeofday(&time1, NULL);
        for (i = 0; i < n_loop; i++)
        {
            status = GdipDrawImage(graphics, (GpImage *)bitmap, 0, 0);
            assert(status == Ok);
        }

        gettimeofday(&time2, NULL);
        printf("bitmap format %#x, graphics format %#x (rotate): %f seconds\n",
                bitmap_fmt, device_fmt, diff_time(&time2, &time1));
    }

    GdipDeleteGraphics(graphics);
    GdipDisposeImage((GpImage *)device_bitmap);
    GdipDisposeImage((GpImage *)bitmap);
}

int main(int argc, char *argv[])
{
    int i, n_loop = 0, flags = 0;
    struct GdiplusStartupInput startup;
    ULONG_PTR token;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "identity"))
            flags |= T_IDENTITY;
        else if (!strcmp(argv[i], "scale"))
            flags |= T_SCALE;
        else if (!strcmp(argv[i], "rotate"))
            flags |= T_ROTATE;
        else if (!strncmp(argv[i], "loop=", 5))
           n_loop = atoi(argv[i] + 5);
    }
    if (!flags) flags = T_IDENTITY | T_SCALE | T_ROTATE;
    if (n_loop <= 0) n_loop = 1000;

    printf("loop count %d\n", n_loop);

    startup.GdiplusVersion = 1;
    startup.DebugEventCallback = NULL;
    startup.SuppressBackgroundThread = 0;
    startup.SuppressExternalCodecs = 0;
    GdiplusStartup(&token, &startup, NULL);

    printf("drawing 1 bpp image on a 1 bpp device\n");
    test_GdipDrawImage(PixelFormat1bppIndexed, PixelFormat1bppIndexed, flags, n_loop);
    printf("drawing 1 bpp image on a 24 bpp device\n");
    test_GdipDrawImage(PixelFormat1bppIndexed, PixelFormat24bppRGB, flags, n_loop);
    printf("drawing 1 bpp image on a 32 bpp device\n");
    test_GdipDrawImage(PixelFormat1bppIndexed, PixelFormat32bppARGB, flags, n_loop);

    printf("drawing 24 bpp image on a 1 bpp device\n");
    test_GdipDrawImage(PixelFormat24bppRGB, PixelFormat1bppIndexed, flags, n_loop);
    printf("drawing 24 bpp image on a 24 bpp device\n");
    test_GdipDrawImage(PixelFormat24bppRGB, PixelFormat24bppRGB, flags, n_loop);
    printf("drawing 24 bpp image on a 32 bpp device\n");
    test_GdipDrawImage(PixelFormat24bppRGB, PixelFormat32bppARGB, flags, n_loop);

    printf("drawing 32 bpp image on a 1 bpp device\n");
    test_GdipDrawImage(PixelFormat32bppARGB, PixelFormat1bppIndexed, flags, n_loop);
    printf("drawing 32 bpp image on a 24 bpp device\n");
    test_GdipDrawImage(PixelFormat32bppARGB, PixelFormat24bppRGB, flags, n_loop);
    printf("drawing 32 bpp image on a 32 bpp device\n");
    test_GdipDrawImage(PixelFormat32bppARGB, PixelFormat32bppARGB, flags, n_loop);

    GdiplusShutdown(token);

    return 0;
}
